from typing import Optional as _Optional

from numpy import float64 as _float64
from pandas import DataFrame as _DataFrame
from pandas import Series as _Series
from pandas import read_csv as _read_csv

from tompy.matrix import describe as _describe
from tompy.matrix import describe_meta as _describe_meta
from tompy.matrix import lr2crdd as _lr2crdd
from tompy.matrix import lr2sr as _lr2sr
from tompy.matrix import lr_metrics as _lr_metrics
from tompy.matrix import lr_sr_metrics as _lr_sr_metrics
from tompy.matrix import lr_sr_metrics_meta as _lr_sr_metrics_meta
from tompy.matrix import portfolio_price as _portfolio_price
from tompy.matrix import price2lr as _price2lr
from tompy.matrix import price2sr as _price2sr
from tompy.matrix import price_metrics as _price_metrics
from tompy.matrix import rolling_max as _rolling_max
from tompy.matrix import rolling_mean as _rolling_mean
from tompy.matrix import rolling_min as _rolling_min
from tompy.matrix import rolling_std as _rolling_std
from tompy.matrix import rolling_sum as _rolling_sum
from tompy.matrix import rolling_var as _rolling_var
from tompy.matrix import shift as _shift
from tompy.matrix import sr2lr as _sr2lr
from tompy.matrix import sr_metrics as _sr_metrics
from tompy.series import rolling_corr as _rolling_corr


def drop_weekends(df: _DataFrame) -> _DataFrame:
    return df[df.index.dayofweek < 5]


def read_csv(
    fpath: str, index_col: _Optional[int] = 0, parse_dates: bool = True
) -> _DataFrame:
    return _read_csv(fpath, index_col=index_col, parse_dates=parse_dates)


def write_csv(df: _DataFrame, fpath: str) -> None:
    df.to_csv(fpath)


def describe(df: _DataFrame) -> _DataFrame:
    return _DataFrame(
        data=_describe(df.values),
        index=_describe_meta().keys(),
        columns=df.columns,
        dtype=_float64,
    )


def shift(df: _DataFrame, lag: int) -> _DataFrame:
    if lag == 0:
        return df
    return _DataFrame(
        data=_shift(df.values, lag),
        index=df.index,
        columns=df.columns,
        dtype=_float64,
    )


def rolling_sum(df: _DataFrame, window: int) -> _DataFrame:
    return _DataFrame(
        data=_rolling_sum(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=_float64,
    )


def rolling_mean(df: _DataFrame, window: int) -> _DataFrame:
    return _DataFrame(
        data=_rolling_mean(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=_float64,
    )


def rolling_std(df: _DataFrame, window: int) -> _DataFrame:
    return _DataFrame(
        data=_rolling_std(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=_float64,
    )


def rolling_var(df: _DataFrame, window: int) -> _DataFrame:
    return _DataFrame(
        data=_rolling_var(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=_float64,
    )


def rolling_min(df: _DataFrame, window: int) -> _DataFrame:
    return _DataFrame(
        data=_rolling_min(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=_float64,
    )


def rolling_max(df: _DataFrame, window: int) -> _DataFrame:
    return _DataFrame(
        data=_rolling_max(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=_float64,
    )


def rolling_corr(df: _DataFrame, window: int, cidx: int = 0) -> _DataFrame:
    s = df[df.columns[cidx]]
    df2 = _DataFrame(index=df.index, columns=df.columns, dtype=_float64)
    for col in df:
        df2[col] = _rolling_corr(s, df[col], window)
    return df2


def rolling_ema(df: _DataFrame, window: int) -> _DataFrame:
    return df.ewm(span=window, adjust=False).mean()


def price2sr(price: _DataFrame) -> _DataFrame:
    return _DataFrame(
        data=_price2sr(price.values),
        index=price.index,
        columns=price.columns,
        dtype=_float64,
    )


def price2lr(price: _DataFrame) -> _DataFrame:
    return _DataFrame(
        data=_price2lr(price.values),
        index=price.index,
        columns=price.columns,
        dtype=_float64,
    )


def price_metrics(
    price: _DataFrame,
    ann_factor: float,
    sqrt_ann_factor: _Optional[float] = None,
    mode: int = 0,
) -> _DataFrame:
    return _DataFrame(
        data=_price_metrics(price.values, ann_factor, sqrt_ann_factor, mode),
        index=_lr_sr_metrics_meta(mode).keys(),
        columns=price.columns,
        dtype=_float64,
    )


def sr2lr(sr: _DataFrame) -> _DataFrame:
    return _DataFrame(
        data=_sr2lr(sr.values),
        index=sr.index,
        columns=sr.columns,
        dtype=_float64,
    )


def sr_metrics(
    sr: _DataFrame,
    ann_factor: float,
    sqrt_ann_factor: _Optional[float] = None,
    mode: int = 0,
) -> _DataFrame:
    return _DataFrame(
        data=_sr_metrics(sr.values, ann_factor, sqrt_ann_factor, mode),
        index=_lr_sr_metrics_meta(mode).keys(),
        columns=sr.columns,
        dtype=_float64,
    )


def lr2sr(lr: _DataFrame) -> _DataFrame:
    return _DataFrame(
        data=_lr2sr(lr.values),
        index=lr.index,
        columns=lr.columns,
        dtype=_float64,
    )


def lr2crdd(lr: _DataFrame) -> dict[str, _DataFrame]:
    lr0_cr, lr0_dd = _lr2crdd(lr.values)
    return {
        "LR_CR": _DataFrame(
            data=lr0_cr, index=lr.index, columns=lr.columns, dtype=_float64
        ),
        "LR_DD": _DataFrame(
            data=lr0_dd, index=lr.index, columns=lr.columns, dtype=_float64
        ),
    }


def lr_metrics(
    lr: _DataFrame,
    ann_factor: float,
    sqrt_ann_factor: _Optional[float] = None,
    mode: int = 0,
) -> _DataFrame:
    return _DataFrame(
        data=_lr_metrics(lr.values, ann_factor, sqrt_ann_factor, mode),
        index=_lr_sr_metrics_meta(mode).keys(),
        columns=lr.columns,
        dtype=_float64,
    )


def lr_sr_metrics(
    lr: _DataFrame,
    sr: _DataFrame,
    ann_factor: float,
    sqrt_ann_factor: _Optional[float] = None,
    mode: int = 0,
) -> _DataFrame:
    return _DataFrame(
        data=_lr_sr_metrics(
            lr.values, sr.values, ann_factor, sqrt_ann_factor, mode
        ),
        index=_lr_sr_metrics_meta(mode).keys(),
        columns=lr.columns,
        dtype=_float64,
    )


def portfolio_price(
    price: _DataFrame,
    weights: _DataFrame,
    fee_rate: float,
) -> _Series:
    price = price[weights.columns]
    price = price.loc[weights.index]
    return _Series(
        data=_portfolio_price(price.values, weights.values, fee_rate),
        index=price.index,
        dtype=_float64,
    )
