from typing import Optional as _Optional

from numpy import full as _full
from numpy import nan as _nan
from numpy import tile as _tile
from pandas import DataFrame as _DataFrame
from pandas import Series as _Series

from tompy.dataframe import portfolio_price as _portfolio_price
from tompy.dataframe import shift as _shift


def rebalancing_weight(
    weights: _DataFrame,
    rebalancing: str,
    lag: int,
) -> _DataFrame:
    if rebalancing == "W":
        idx = _DataFrame(index=weights.index)
        idx["W"] = idx.index.dayofweek
        idx["NA"] = idx["W"].diff(-1).fillna(0) <= 0
        weights = weights.copy()
        weights.loc[idx["NA"]] = _nan
    elif rebalancing == "M":
        idx = _DataFrame(index=weights.index)
        idx["M"] = idx.index.month
        idx["NA"] = idx["M"].diff(-1).fillna(0) == 0
        weights = weights.copy()
        weights.loc[idx["NA"]] = _nan
    return _shift(weights, lag)


def crp_ew(price: _DataFrame, tickers: list[str]) -> _DataFrame:
    price = price[tickers].dropna()
    nr = price.shape[0]
    nc = len(tickers)
    ew = 1.0 / nc
    m = _full((nr, nc), ew)
    return _DataFrame(data=m, index=price.index, columns=tickers)


def crp(price: _DataFrame, ticker_weights: dict[str, float]) -> _DataFrame:
    tickers = list(ticker_weights.keys())
    weights = list(ticker_weights.values())
    price = price[tickers].dropna()
    nr = price.shape[0]
    m = _tile(weights, (nr, 1))
    return _DataFrame(data=m, index=price.index, columns=tickers)


class Base:
    def __init__(
        self,
        price: _DataFrame,
    ) -> None:
        self.__price = price

    @property
    def _price(self) -> _DataFrame:
        return self.__price

    def portfolio_weight(self) -> _DataFrame:
        raise NotImplementedError

    def rebalancing_weight(
        self,
        weights: _Optional[_DataFrame] = None,
        rebalancing: str = "D",
        lag: int = 1,
    ) -> _DataFrame:
        if weights is None:
            weights = self.portfolio_weight()
        return rebalancing_weight(weights, rebalancing, lag)

    def portfolio_price(self, weights: _DataFrame, fee_rate: float) -> _Series:
        return _portfolio_price(self._price, weights, fee_rate)


class CRP_EW(Base):
    def __init__(
        self,
        price: _DataFrame,
        tickers: list[str],
    ) -> None:
        super().__init__(price)
        self.__tickers = list(dict.fromkeys(tickers))

    @property
    def _tickers(self) -> list[str]:
        return self.__tickers

    def portfolio_weight(self) -> _DataFrame:
        return crp_ew(self._price, self._tickers)


class CRP(Base):
    def __init__(
        self,
        price: _DataFrame,
        ticker_weights: dict[str, float],
    ) -> None:
        super().__init__(price)
        self.__ticker_weights = ticker_weights

    @property
    def _ticker_weights(self) -> dict[str, float]:
        return self.__ticker_weights

    def portfolio_weight(self) -> _DataFrame:
        return crp(self._price, self._ticker_weights)
